<?php
## Gladius Database Engine
# @author legolas558
# @version 0.1.68
# Licensed under GNU General Public License (GPL)
#
#
# Flatfile php database engine compliant with a subset of intermediate SQL92
#
# TODO:
# - arithmetical subexpressions (parentheses not recognized)
# - consistent INDEX implementation
# - auto_increment must be after not null
# - consistent AUTO_INCREMENT / SERIAL datatype implementation
# - HAVING and GROUP BY implementation (v0.2)
# - JOINs (v0.2)
# NOTES:
# - proper statement termination needs enforcement checks
# - pool can be cached on disk and updated when data changes

include 'gladius_const.php';

function gladius_load($fname) {
	$rd = substr(file_get_contents($fname), strlen(_G__PHP_HEAD));
	if ($rd{0} !== 'a')
		return unserialize(gzuncompress($rd));
	return unserialize($rd);
}

// load a compressed/plain file taking care of shared read locks
function gladius_load2($fn) {
	$f = fopen($fn, 'rb');
	$ofs = strlen(_G__PHP_HEAD);
	flock($f, LOCK_SH);
	fseek($f, $ofs);
	$rd = fread($f, filesize($fn)-$ofs);
	flock($f, LOCK_UN);
	fclose($f);
	if ($rd{0} !== 'a')
		return unserialize(gzuncompress($rd));
	return unserialize($rd);
}

function gladius_save($fname, &$var) {
	$f = fopen($fname, 'wb');
	flock($f, LOCK_EX);
	fwrite($f, _G__PHP_HEAD);
	$s = serialize($var);
	if (strlen($s) > _G__COMPRESSION_THRESHOLD)
		fwrite($f, gzcompress($s, 3));
	else
		fwrite($f, $s);
	flock($f, LOCK_UN);
	fclose($f);
}

// custom data decoder/encoder functions
function gladius_datetime_enc($data, &$dest) {
	if ($data === null) {
		$dest = null;
		return true;
	}

	if (!preg_match('/((\\d\\d\\d\\d)-(\\d\\d)-(\\d\\d)\\s+)?((\\d\\d):(\\d\\d)(:(\\d\\d))?)?/A',
					$data, $m))
		return false;

	if ($m[0] === '')
		return false;
	$dest = mktime((int)@$m[6], (int)@$m[7], (int)@$m[9], (int)@$m[3], (int)@$m[4], (int)@$m[2]);

	return true;
}

function gladius_time_enc($data, &$dest) {
	return gladius_datetime_enc($data, $dest);
}

function gladius_date_enc($data, &$dest) {
	return gladius_datetime_enc($data, $dest);
}

function gladius_datetime_dec($data, &$dest) {
	if ($data === null) {
		$dest = null;
		return true;
	}
	$dest = strftime(_G__DATETIME_FORMAT, $data);
	return true;
}

function gladius_time_dec($data, &$dest) {
	if ($data === null) {
		$dest = null;
		return true;
	}
	$dest = strftime(_G__TIME_FORMAT, $data);
	return true;
}

function gladius_date_dec($data, &$dest) {
	if ($data === null) {
		$dest = null;
		return true;
	}
	$dest = strftime(_G__DATE_FORMAT, $data);
	return true;
}

function equ_comp(&$lh, &$rh) {
	if (gettype($lh) != gettype($rh))
		return (strcmp($lh, $rh) == 0);
	return ($lh == $rh);
}

function array_set($keys, $value) {
	$r=array();
	foreach ($keys as $key)
		$r[$key] = $value;
	return $r;
}

// base class with error handling features
class Gladius_Core {
	var $errstr;
	var $errno = _G__SUCCESS;
	var $_command;
	var $offset;

	function _log($msg) {
	// enable the below line to show errors
/*		echo '<p style="font-family: lucida, tahoma; font-weight:bold; color: red;">[GLADIUS DB] '.$msg.'</p>';	 */
/*		trigger_error($msg, E_USER_NOTICE);	*/
	}

	function _error($errno) {
		global $GLADIUS_ERRORS;
		$this->errno = $errno;

		if (func_num_args()>1) {
			$args = array_slice(func_get_args(), 1);
			$this->errstr=call_user_func_array('sprintf', array_merge(array($GLADIUS_ERRORS[$errno]), $args));
		} else
			$this->errstr=$GLADIUS_ERRORS[$errno];

		if ($this->_command)
			$this->errstr .= ' in command '.$this->_command;
		$this->errstr.=' at offset '.$this->offset;

		if ($dbg = @debug_backtrace())
			$dbginfo = 'line '.$dbg[0]['line'].', file '.basename($dbg[0]['file']).' ';
		else $dbginfo = '';

		$this->_log($dbginfo.$this->errstr);
	}

} // class Gladius_Core

class Gladius extends Gladius_Core {

	var $database;		// the database folder path
	var $_command;
	var $fetch_mode;
	var $insert_id;
	var $affected_rows;
	var $schemas;
	var $pool;			// $pool['table']['column'] = array($value1, $value2, ...);

	function Gladius($db = null) {
		$this->_reset();
		$this->fetch_mode = GLADIUS_FETCH_DEFAULT;

		if (isset($db))
			$this->SelectDB($db);
	}

	function _reset() {
		$this->pool = array();
		$this->affected_rows = 0;
		$this->schemas = null;
		$this->_command = '';
		$this->insert_id = -1;
	}

	// deprecated, will be removed in v0.2
	function Select($dbname) {
		return $this->SelectDB($dbname);
	}

	/* public */ function SelectDB($dbname) {
		global $GLADIUS_DB_ROOT;
		$path = $GLADIUS_DB_ROOT.$dbname;
		if (!is_dir($path)) {
			$this->_error(_G__DB_NOT_FOUND, $dbname);
			return false;
		}
		$old_db = $this->database;
		if ($dbname{strlen($dbname)-1}!='/')
			$this->database = $path.'/';
		else
			$this->database = $path;
		if ($old_db != $this->database)
			$this->_reset();
		return true;
	}

	/* public */ function Escape($string, $magic_quotes) {
		return str_replace("'", "''", $string);
	}

	/* public */ function _unescape($s) {
		return str_replace("''", "'", $s);
	}

	function _recon_sql_type($type, &$fo) {
		$type = strtolower($type);

		switch ($type) {
			case 'int':		// not in SQL92 standard
			case 'integer':
			case 'smallint':
			case 'bigint':
			case 'tinyint':	// not in SQL92 standard
				$fo->type = 'int';
				$fo->numeric = true;
				return true;
		}

		switch ($type) {
			case 'char':
			case 'varchar':
				$fo->type = 'varchar';
				$fo->numeric = false;
				return true;
		}

		switch ($type) {
			case 'numeric':
			case 'decimal':
			case 'float':
			case 'real':
			case 'double':	// not in SQL92 standard
//			case 'double precision': // not yet implemented
				$fo->type = 'double';
				$fo->numeric = true;
				return true;
			case 'text':
				$fo->type = 'text';
				$fo->numeric = false;
			return true;
			case 'date':
			case 'time':
			case 'datetime': // not in SQL92 standard
			case 'timestamp':
				$fo->type = $type;
				$fo->numeric = true;
			return true;
		}

		return false;
	}

	function _get_hand_imm(&$sql) {
		if (preg_match('/\\s*'._G__STRING.'/A', $sql, $m, PREG_OFFSET_CAPTURE, $this->offset)) {
			$this->offset = $m[0][1]+strlen($m[0][0]);
			return new value((string)substr(trim($m[0][0]),1,-1));
		}
		if (preg_match('/\\s*([\\+-]?\\d+\\.\\d+)/A', $sql, $m, PREG_OFFSET_CAPTURE, $this->offset)) {
			$this->offset = $m[0][1]+strlen($m[0][0]);
			return new value((double)$m[1][0]);
		}
		if (preg_match('/\\s*([\\+-]?\\d+)/A', $sql, $m, PREG_OFFSET_CAPTURE, $this->offset)) {
			$this->offset = $m[0][1]+strlen($m[0][0]);
			return new value((int)$m[1][0]);
		}
		if (preg_match('/\\s*(\\w+)\\s*\\(/A', $sql, $m, PREG_OFFSET_CAPTURE, $this->offset)) {
			$this->offset = $m[0][1]+strlen($m[0][0]);
			$subexpr = $this->_get_hand_expr($sql, true);
			return new func($m[1][0], $subexpr);
		}
		return false;
	}

	function _retrieve_value($value, &$fo, &$dest) {
		
		if (strtoupper($value)=='NULL') {
			if ($fo->not_null) {
				$this->_error(_G__FIELD_NOT_NULL, $fo->name);
				return false;
			}
			$dest = $fo->default;
			return true;
		}

		if (strlen($value)) {
			if (($value{0}=="'") && ($value{strlen($value)-1}=="'"))
				$value = substr($value, 1, -1);
		}

		if ($fo->numeric) {

			switch ($fo->type) {
				case 'int':
					if (!is_numeric($value)) {
						$this->_error(_G__FIELD_TYPE_MISMATCH, $fo->name, $fo->type);
						return false;
					}

/*					if (isset($fo->max_length)) {
						if ($value>pow(2, $fo->max_length)) {
							$this->_error(_G__EXCEED_MAX_LENGTH,
										  $fo->name, $value);
							return false;
						}
					}	*/
					$dest = (int)$value;
					return true;
				break;
				case 'double':
					if (!is_numeric($value)) {
						$this->_error(_G__FIELD_TYPE_MISMATCH, $fo->name, $fo->type);
						return false;
					}
					if (isset($fo->max_length)) {
						$lcl = localeconv();
						$p = strpos($value, $lcl['decimal_point']);
						if ($p===false)	$p=strlen($value);
						if ($p > $fo->max_length) {
							$this->_error(_G__EXCEED_MAX_LENGTH,
										  $fo->name, $value);
							return false;
						}
					}
					$dest = (double)$value;
					return true;
				break;
				case 'timestamp':
					$dest = (int)$value;
					return true;
				case 'date':
				case 'time':
				case 'datetime':
					$encfunc = 'gladius_'.$fo->type.'_enc';
					if (!$encfunc($value, $ignored)) {
						$this->_error(_G__CONVERSION_ERROR, $fo->type);
						return false;
					}
					$dest = (int)$value;
					return true;
				default:
					die('GRTE: invalid numeric type '.$fo->type);
			}
			$dest =  $v;
			return true;
		} else {
			$value = $this->_unescape($value);
			switch ($fo->type) {
				case 'varchar':
				case 'text':
					if (isset($fo->max_length)) {
						if (strlen($value) > $fo->max_length) {
							$this->_error(_G__EXCEED_MAX_LENGTH,
										  $fo->name, substr($value,0,$fo->max_length).'...');
							return false;
						}
					}
					$dest = $value;
					return true;
				default:
					die('GRTE: invalid string type '.$fo->type);
			}
		}
		return false;
	}

	function _close_ok(&$m, &$ok) {
		switch ($m[count($m)-1][0]) {
			case ')':
				$ok = true;
			case ',':
				return true;
		}
		$this->_error(_G__MALFORMED_SQL);
		return false;
	}

	function _append_record($table, &$fo, $data) {
		$path = $this->database.$table.'.'.$fo->name.'.php';
		if (!isset($this->pool[$table][$fo->name]))
			$this->pool[$table][$fo->name] = gladius_load($path);
		$column =& $this->pool[$table][$fo->name];
		if ($fo->unique) {
			if (in_array($data, $column)) {
				$this->_error(_G__UNIQUE_VIOLATED, $fo->name);
				return false;
			}
		}

		global $GLADIUS_CUSTOM_DATA;
		if (in_array($fo->type, $GLADIUS_CUSTOM_DATA)) {
			$encfunc = 'gladius_'.$fo->type.'_enc';
			if (!$encfunc($data, $converted)) {
				$this->_error(_G__CONVERSION_ERROR, $fo->type);
				return false;
			}
		} else
			$converted =& $data;
		$column[] =& $converted;
		// save data to file
		gladius_save($path, $column);
		if ($fo->indexed)
			$this->_update_order($fo, $column, $converted);
		return true;
	}

	function _update_order(&$fo, &$column, &$data) {
		// reload columns in their ascending order
		$c = count($fo->ordered);
		$p_i = 0;
		$done = false;
		for ($i=0;$i<$c;$i++) {
			if ($data < $column[$fo->ordered[$i]]) {
//				array_splice($fo->ordered, $p_i, 0, $c);
				$ordered_col = array_slice($fo->ordered, 0, $p_i);
				$ordered_col[] = $c;
				$fo->ordered = array_merge($ordered_col, array_slice($fo->ordered, $p_i));
				$done = true;
				break;
			}
			$p_i = $i;
		}
		if (!$done)
			$fo->ordered[] = $c;
	}

	function _create_table($table, &$schema) {
		$empt = array();
		foreach ($schema['FIELD_DEF'] as $fo) {
			gladius_save($this->database.$table.'.'.$fo->name.'.php', $empt);
			$this->pool[$table][$fo->name] = array();
		}
	}

	function _parse_from(&$sql, &$from) {
		while (preg_match('/\\s*(\\w+)(\\s+(as\\s+)?(\\w+))?(\\s*,)?/iA',
				$sql, $m, PREG_OFFSET_CAPTURE, $this->offset)) {

			$c = count($m);
			if ($c > 2) {
				$alias = strtoupper($m[4][0]);
				switch ($alias) {
					case 'WHERE':
					case 'ORDER':
					case 'GROUP':
					case 'LIMIT':
						if (!empty($m[3][0])) {
							$this->_error(_G__INVALID_ALIAS, $m[2][0]);
							return false;
						}
						$from[] = array(strtoupper($m[1][0]), '');
						$this->offset = $m[4][1];
						break 2;
				}
			} else
				$alias = '';

			$src = strtoupper($m[1][0]);
			if (in_array($src, $from, true)) {
				$this->_error(_G__DUPLICATE_SOURCE, $src);
				return false;
			}

			$from[] = array($src, $alias);

			if ($m[$c-1][0]==',') {
				$this->offset = $m[$c-1][1]+1;
			} else {
				$this->offset = $m[$c-1][1]+strlen($m[$c-1][0]);
				break;
			}
		}

		if (empty($from)) {
			$this->_error(_G__NO_SOURCE);
			return false;
		}
		return true;
	}

	/* public */ function Query($sql) {
		if (!preg_match('/\\s*(\\w+)\\s*/i', $sql, $m, PREG_OFFSET_CAPTURE))
			return false;

		$this->_command = strtoupper($m[1][0]);
		switch ($this->_command) {

			case 'CREATE':
				return include 'gladius_create.php';

			case 'DROP':
				return include 'gladius_drop.php';

			case 'DELETE':
				include_once 'gladius_sqlops.php';
				include_once 'gladius_rs.php';
				return include 'gladius_delete.php';

			case 'INSERT':
				include_once 'gladius_sqlops.php';
				include_once 'gladius_rs.php';
				return include 'gladius_insert.php';

			case 'UPDATE':
				include_once 'gladius_sqlops.php';
				include_once 'gladius_rs.php';
				return include 'gladius_update.php';

			case 'SELECT':
				include_once 'gladius_sqlops.php';
				include_once 'gladius_rs.php';
				$this->offset = $m[1][1]+6;
				return $this->do_select($sql, false);

			case 'USE':
				$this->offset = $m[1][1]+6;
				if (!preg_match('/\\s+(\\w+)/A', $sql, $m,
								PREG_OFFSET_CAPTURE, $this->offset)) {
					$this->_error(_G__MALFORMED_SQL);
					return false;
				}
				$this->offset = $m[1][1]+strlen($m[1][0]);
				return $this->SelectDB($m[1][0]);

			case 'TRUNCATE':

			default:
				$this->_command = null;
				$this->_error(_G__UNRECON_SQL, strtoupper($m[1][0]));
				return false;
		}
	}

	function do_select(&$sql, $subquery) {
		return include 'gladius_select.php';
	}

	// load needed tables in memory and create a new resultset
	function _create_rs($from) {
		$rs = new Gladius_Resultset($this);

		$rs->sources = array();
		$rs->tables = array();
		foreach ($from as $table) {
			if (empty($table[1]))
				$alias = $table[0];
			else
				$alias = $table[1];
			if (!$this->_hold_schema($table[0]))
				return false;
			$rs->sources[$alias] =& $this->schemas[$table[0]];
/*			// create records cache
			$rs->sources[$alias]['RECORDS'] = array_set(
				array_values($rs->sources[$alias]['RECORD_IDS']),
				null
			);
			$rs->sources[$alias]['CUR_ROW'] = 0;	*/
			reset($rs->sources[$alias]['RECORD_IDS']);
			if (!in_array($table[0], $rs->tables))
				$rs->tables[] = $table[0];
			if (!empty($table[1]))
				$rs->sources[$table[0]] =& $rs->sources[$table[1]];
		}
		$rs->_reset_tables();
		return $rs;
	}

	function _hold_schema($table) {
		if (!isset($this->schemas[$table])) {
			$schema_src = $this->database.$table.'.php';
			if (!file_exists($schema_src)) {
				$this->_error(_G__TABLE_NOT_FOUND, $table);
				return false;
			}
			$this->schemas[$table] = gladius_load($schema_src);
		}
		return true;
	}

	function _hold_column($table, $column) {
		if (!isset($this->pool[$table][$column])) {
			$column_src = $this->database.$table.'.'.$column.'.php';
			$this->pool[$table][$column] = gladius_load($column_src);
		}
	}

	function _success($affected) {
		global $GLADIUS_ERRORS;
		$this->errstr = sprintf($GLADIUS_ERRORS[_G__SUCCESS], $this->_command, $affected, (($affected == 1) ? '': 's'));
		$this->affected_rows = $affected;
		$this->errno = _G__SUCCESS;
	}


}	// Gladius class declaration ends

?>