<?php
## Secure initialization functions
# @author legolas558
#
# experimental functions to allow a better security layer inside Drake
# these functions tend to fill the 'controller' layer which was totally missing
# functions are supplied in different flavours to ease coding

	define('__RAW', 0);				// no changes will be performed to input
	define('__SQL', 1);				// en/decoded to be safe for the current database
	define('__XHTML', 2);			// html en/decoded
	define('__PATH', 4);			// input will be safe for path usage
	define('__NUM', 16);			// positive integer number or null
	define('__RFC2045', 32);		// RFC2045 valid header (<76 in length and no newlines)
	define('__ARR', 64);			// Array type (to be recursively scanned)
	define('__ARR0', 128);		// Same as above, but only the first element will be considered
	/*NOTE: __ARR0 can be considered transitional (it is preferred not to use it)
		while a new approach is researched for single value arrays */
	define('__NOHTML', 256);		// strip HTML tags

	function croplen($data, $max) {
		if ($max) {
			if (strlen($data)>$max)
				$data = substr($data, 0, $max);
		}
	}

	/**
	 * encode a string for SQL statements inclusion
	 * note: global slashes quoting (gp magic quotes) is not
	 * inherent to this function
	 * @return The \var $data variable with proper quoting
	 * done through the adoDB qstr() function
	 */
	function sql_encode($data, $max=null) {
    	global $conn;
		croplen($data, $max);
	    return $conn->Quote($data, false);
	}

	function xhtml_encode($data, $max) {
		global $d_root, $my_kses;
		require_once($d_root.'includes/xhtml.php');
		croplen($data, $max);
		return $my_kses->Parse($data);
	}

	function num_get($data, $def=null) {
		if (!ctype_digit($data))
			return $def;
		else
			return (int)$data;
	}

	function path_safe($data) {
		$data = str_replace(array('/','\\'), '_', $data);
		if ($data{0}=='.') $data{0}='_';
		return strtolower($data);
	}

	/**
	 * prepare input for sql usage.
	 * @return The variable input prepared by sql_encode for sql
	 * usage, not including quoting. Quoting is inherent to the SQL statement
	 */
	function in_sql($name, $arr = null, $def = null, $max = null) {
		if ($arr===null)
			$arr =& $_REQUEST;
		if (null === ($data = @$arr[$name])) return $def;
		//s2: this should now be handled by the adodb driver
		//L: global quoting settings are not inherent to sql_encode()
		raw_fix($data);
		return sql_encode($data, $max);
	}

	function in_html($name, $arr = null,$def = null, $max = null) {
		$data = in_raw($name, $arr, $def, $max);
		return xhtml_encode($data, $max);
	}

	function in_num($name, $arr = null,$def = null) {
		if ($arr===null)
			$arr =& $_REQUEST;
		if (!isset($arr[$name]))	return $def;
		$data = $arr[$name];
		// raw_fix($data) is not needed here
		return num_get($data, $def);
	}

	function in_checkbox($name, $arr = null, $def = 0) {
		if ($arr===null)
			$arr =& $_REQUEST;
		if (!isset($arr[$name]))	return $def;
		$data = $arr[$name];
		if (!$data)	return 0;	// will evaluate '0', 'false' and ''
		// raw_fix($data) is not needed here
		if ($data!='off')
			return 1;
		return 0;
	}

	function in_raw($name, $arr = null, $def = null, $max = null) {
		if ($arr===null)
			$arr =& $_REQUEST;
		if (null === ($data = @$arr[$name])) return $def;
		raw_fix($data);
		croplen($data, $max);
		return $data;
	}

	function in_path($name, $arr = null, $def = null, $max = null) {
		if ($arr===null)
			$arr =& $_REQUEST;
		if (null === ($data = @$arr[$name])) return $def;
		raw_fix($data);
		$data = path_safe($data);
		croplen($data, $max);
		return $data;
	}

	function rfc2045_safe($data) {
		$n = explode("\n", $data, 2);
		$l = explode("\r", $n[0], 2);
		return $l[0];
	}

	function in_arr($name, $arr = null, $def = null, $flags=null, $max = null) {
		if ($arr===null)
			$arr =& $_REQUEST;
		if (null === ($data = @$arr[$name]))
			return $def;
		// raw_fix($data) is not needed here since it is executed by in_validation()
		if (!is_array($data))
			return $def;

		$c=count($data);
		for($i=0;$i<$c;$i++)
			in_validation($data[$i], $flags , $def, $max);
		return $data;
	}

	if (get_magic_quotes_gpc()) {
		function raw_fix(&$data) {
			$data = stripslashes($data);
		}
	} else {
		function raw_fix(&$data) {}
	}

	function in_validation(&$data, $flags, $def, $max) {
		if (is_array($data))	// disallow nested arrays
			return $def;
		raw_fix($data);
		if ($flags & __NUM) {
			$data = num_get($data);
			if ($data===null) return $def;
			return $data;
		}

		if ($flags & __XHTML)
			$data = xhtml_encode($data, $max);
		if ($flags & __NOHTML)
			$data = htmlentities($data, $max);
		if ($flags & __SQL)
			$data = sql_encode($data, $max);
		if ($flags & __PATH)
			$data = path_safe($data);
		if ($flags & __RFC2045)
			$data = rfc2045_safe($data);
		croplen($data, $max);
	}

	// prepare a variable for input from the (wild) external world
	function in($name, $flags = 0, &$arr, $def = null, $max = null) {
		if ($arr===null)
			$arr =& $_REQUEST;
		if (null === ($data = @$arr[$name]))
			return $def;
		// raw_fix($data) is not needed here since it is executed by in_validation()
		if ($flags & __ARR0) {
			if (!isset($data[0]))
				return $def;
			in_validation($data[0], $flags ^ __ARR0 , $def, $max);
			return $data[0];
		}
		if ($flags & __ARR) {
			if (!is_array($data))
				return $def;

			$c=count($data);
			for($i=0;$i<$c;$i++)	// referencing pointers to array items (through '&') is allowed only on PHP5+
				in_validation($data[$i], $flags ^ __ARR , $def, $max);
			return $data;
		}

		in_validation($data, $flags, $def, $max);
		return $data;
	}

?>